#!/usr/env python
'''
Tools for histogram
'''

from .parzen_estimator import ParzenEstimator
from .stats import Stats
from .findpeaks import FindPeaks
from .gaussfit import GaussFit
from .gaussrealfit import GaussRealFit
